@@ -1,13 +1,7 @@
-from __future__ import unicode_literals
 import re
 
-import html5lib
-from html5lib.filters.base import Filter
-from html5lib.filters.sanitizer import allowed_protocols
-from html5lib.serializer import HTMLSerializer
-
 from bleach import callbacks as linkify_callbacks
-from bleach.encoding import force_unicode
+from bleach import html5lib_shim
 from bleach.utils import alphabetize_attributes
 
 
@@ -33,17 +27,17 @@ TLDS = """ac ad ae aero af ag ai al am an ao aq ar arpa as asia at au aw ax az
 TLDS.reverse()
 
 
-def build_url_re(tlds=TLDS, protocols=allowed_protocols):
+def build_url_re(tlds=TLDS, protocols=html5lib_shim.allowed_protocols):
     """Builds the url regex used by linkifier
 
-   If you want a different set of tlds or allowed protocols, pass those in
-   and stomp on the existing ``url_re``::
+    If you want a different set of tlds or allowed protocols, pass those in
+    and stomp on the existing ``url_re``::
 
-       from bleach import linkifier
+        from bleach import linkifier
 
-       my_url_re = linkifier.build_url_re(my_tlds_list, my_protocols)
+        my_url_re = linkifier.build_url_re(my_tlds_list, my_protocols)
 
-       linker = LinkifyFilter(url_re=my_url_re)
+        linker = LinkifyFilter(url_re=my_url_re)
 
     """
     return re.compile(
@@ -53,28 +47,50 @@ def build_url_re(tlds=TLDS, protocols=allowed_protocols):
         (?:[/?][^\s\{{\}}\|\\\^\[\]`<>"]*)?
             # /path/zz (excluding "unsafe" chars from RFC 1738,
             # except for # and ~, which happen in practice)
-        """.format('|'.join(protocols), '|'.join(tlds)),
-        re.IGNORECASE | re.VERBOSE | re.UNICODE)
+        """.format(
+            "|".join(sorted(protocols)), "|".join(sorted(tlds))
+        ),
+        re.IGNORECASE | re.VERBOSE | re.UNICODE,
+    )
 
 
 URL_RE = build_url_re()
 
 
-PROTO_RE = re.compile(r'^[\w-]+:/{0,3}', re.IGNORECASE)
+PROTO_RE = re.compile(r"^[\w-]+:/{0,3}", re.IGNORECASE)
 
 
-EMAIL_RE = re.compile(
-    r"""(?<!//)
-    (([-!#$%&'*+/=?^_`{}|~0-9A-Z]+
-        (\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*  # dot-atom
-    |^"([\001-\010\013\014\016-\037!#-\[\]-\177]
-        |\\[\001-\011\013\014\016-\177])*"  # quoted-string
-    )@(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6})  # domain
-    """,
-    re.IGNORECASE | re.MULTILINE | re.VERBOSE)
+def build_email_re(tlds=TLDS):
+    """Builds the email regex used by linkifier
+
+    If you want a different set of tlds, pass those in and stomp on the existing ``email_re``::
+
+        from bleach import linkifier
+
+        my_email_re = linkifier.build_email_re(my_tlds_list)
+
+        linker = LinkifyFilter(email_re=my_url_re)
+
+    """
+    # open and closing braces doubled below for format string
+    return re.compile(
+        r"""(?<!//)
+        (([-!#$%&'*+/=?^_`{{}}|~0-9A-Z]+
+            (\.[-!#$%&'*+/=?^_`{{}}|~0-9A-Z]+)*  # dot-atom
+        |^"([\001-\010\013\014\016-\037!#-\[\]-\177]
+            |\\[\001-\011\013\014\016-\177])*"  # quoted-string
+        )@(?:[A-Z0-9](?:[A-Z0-9-]{{0,61}}[A-Z0-9])?\.)+(?:{0}))  # domain
+        """.format(
+            "|".join(tlds)
+        ),
+        re.IGNORECASE | re.MULTILINE | re.VERBOSE,
+    )
 
 
-class Linker(object):
+EMAIL_RE = build_email_re()
+
+
+class Linker:
     """Convert URL-like strings in an HTML fragment to links
 
     This function converts strings that look like URLs, domain names and email
@@ -88,8 +104,16 @@ class Linker(object):
     situations due to crazy text.
 
     """
-    def __init__(self, callbacks=DEFAULT_CALLBACKS, skip_tags=None, parse_email=False,
-                 url_re=URL_RE, email_re=EMAIL_RE):
+
+    def __init__(
+        self,
+        callbacks=DEFAULT_CALLBACKS,
+        skip_tags=None,
+        parse_email=False,
+        url_re=URL_RE,
+        email_re=EMAIL_RE,
+        recognized_tags=html5lib_shim.HTML_TAGS,
+    ):
         """Creates a Linker instance
 
         :arg list callbacks: list of callbacks to run when adjusting tag attributes;
@@ -105,6 +129,9 @@ class Linker(object):
 
         :arg re email_re: email matching regex
 
+        :arg list-of-strings recognized_tags: the list of tags that linkify knows about;
+            everything else gets escaped
+
         :returns: linkified text as unicode
 
         """
@@ -114,15 +141,20 @@ class Linker(object):
         self.url_re = url_re
         self.email_re = email_re
 
-        self.parser = html5lib.HTMLParser(namespaceHTMLElements=False)
-        self.walker = html5lib.getTreeWalker('etree')
-        self.serializer = HTMLSerializer(
-            quote_attr_values='always',
+        # Create a parser/tokenizer that allows all HTML tags and escapes
+        # anything not in that list.
+        self.parser = html5lib_shim.BleachHTMLParser(
+            tags=recognized_tags,
+            strip=False,
+            consume_entities=True,
+            namespaceHTMLElements=False,
+        )
+        self.walker = html5lib_shim.getTreeWalker("etree")
+        self.serializer = html5lib_shim.BleachHTMLSerializer(
+            quote_attr_values="always",
             omit_optional_tags=False,
-
             # linkify does not sanitize
             sanitize=False,
-
             # linkify alphabetizes
             alphabetical_attributes=False,
         )
@@ -134,11 +166,14 @@ class Linker(object):
 
         :returns: linkified text as unicode
 
+        :raises TypeError: if ``text`` is not a text type
+
         """
-        text = force_unicode(text)
+        if not isinstance(text, str):
+            raise TypeError("argument must be of text type")
 
         if not text:
-            return u''
+            return ""
 
         dom = self.parser.parseFragment(text)
         filtered = LinkifyFilter(
@@ -152,7 +187,7 @@ class Linker(object):
         return self.serializer.render(filtered)
 
 
-class LinkifyFilter(Filter):
+class LinkifyFilter(html5lib_shim.Filter):
     """html5lib filter that linkifies text
 
     This will do the following:
@@ -165,8 +200,16 @@ class LinkifyFilter(Filter):
     This filter can be used anywhere html5lib filters can be used.
 
     """
-    def __init__(self, source, callbacks=None, skip_tags=None, parse_email=False,
-                 url_re=URL_RE, email_re=EMAIL_RE):
+
+    def __init__(
+        self,
+        source,
+        callbacks=DEFAULT_CALLBACKS,
+        skip_tags=None,
+        parse_email=False,
+        url_re=URL_RE,
+        email_re=EMAIL_RE,
+    ):
         """Creates a LinkifyFilter instance
 
         :arg TreeWalker source: stream
@@ -230,22 +273,22 @@ class LinkifyFilter(Filter):
         # linkify callables.
         #
         # I'm not really sure how else to support that ``_text`` fauxttribute and
-        # maintain some modicum of backwards compatability with previous versions
+        # maintain some modicum of backwards compatibility with previous versions
         # of Bleach.
 
         out = []
         for token in token_list:
-            token_type = token['type']
-            if token_type in ['Characters', 'SpaceCharacters']:
-                out.append(token['data'])
+            token_type = token["type"]
+            if token_type in ["Characters", "SpaceCharacters"]:
+                out.append(token["data"])
 
-        return u''.join(out)
+        return "".join(out)
 
     def handle_email_addresses(self, src_iter):
         """Handle email addresses in character tokens"""
         for token in src_iter:
-            if token['type'] == 'Characters':
-                text = token['data']
+            if token["type"] == "Characters":
+                text = token["data"]
                 new_tokens = []
                 end = 0
 
@@ -253,39 +296,41 @@ class LinkifyFilter(Filter):
                 for match in self.email_re.finditer(text):
                     if match.start() > end:
                         new_tokens.append(
-                            {u'type': u'Characters', u'data': text[end:match.start()]}
+                            {"type": "Characters", "data": text[end : match.start()]}
                         )
 
                     # Run attributes through the callbacks to see what we
                     # should do with this match
                     attrs = {
-                        (None, u'href'): u'mailto:%s' % match.group(0),
-                        u'_text': match.group(0)
+                        (None, "href"): "mailto:%s" % match.group(0),
+                        "_text": match.group(0),
                     }
                     attrs = self.apply_callbacks(attrs, True)
 
                     if attrs is None:
                         # Just add the text--but not as a link
                         new_tokens.append(
-                            {u'type': u'Characters', u'data': match.group(0)}
+                            {"type": "Characters", "data": match.group(0)}
                         )
 
                     else:
                         # Add an "a" tag for the new link
-                        _text = attrs.pop(u'_text', '')
+                        _text = attrs.pop("_text", "")
                         attrs = alphabetize_attributes(attrs)
-                        new_tokens.extend([
-                            {u'type': u'StartTag', u'name': u'a', u'data': attrs},
-                            {u'type': u'Characters', u'data': force_unicode(_text)},
-                            {u'type': u'EndTag', u'name': 'a'}
-                        ])
+                        new_tokens.extend(
+                            [
+                                {"type": "StartTag", "name": "a", "data": attrs},
+                                {"type": "Characters", "data": str(_text)},
+                                {"type": "EndTag", "name": "a"},
+                            ]
+                        )
                     end = match.end()
 
                 if new_tokens:
                     # Yield the adjusted set of tokens and then continue
                     # through the loop
                     if end < len(text):
-                        new_tokens.append({u'type': u'Characters', u'data': text[end:]})
+                        new_tokens.append({"type": "Characters", "data": text[end:]})
 
                     for new_token in new_tokens:
                         yield new_token
@@ -300,17 +345,17 @@ class LinkifyFilter(Filter):
         This accounts for over-eager matching by the regex.
 
         """
-        prefix = suffix = ''
+        prefix = suffix = ""
 
         while fragment:
             # Try removing ( from the beginning and, if it's balanced, from the
             # end, too
-            if fragment.startswith(u'('):
-                prefix = prefix + u'('
+            if fragment.startswith("("):
+                prefix = prefix + "("
                 fragment = fragment[1:]
 
-                if fragment.endswith(u')'):
-                    suffix = u')' + suffix
+                if fragment.endswith(")"):
+                    suffix = ")" + suffix
                     fragment = fragment[:-1]
                 continue
 
@@ -320,21 +365,21 @@ class LinkifyFilter(Filter):
             #
             #     "i looked at the site (at http://example.com)"
 
-            if fragment.endswith(u')') and u'(' not in fragment:
+            if fragment.endswith(")") and "(" not in fragment:
                 fragment = fragment[:-1]
-                suffix = u')' + suffix
+                suffix = ")" + suffix
                 continue
 
             # Handle commas
-            if fragment.endswith(u','):
+            if fragment.endswith(","):
                 fragment = fragment[:-1]
-                suffix = u',' + suffix
+                suffix = "," + suffix
                 continue
 
             # Handle periods
-            if fragment.endswith(u'.'):
+            if fragment.endswith("."):
                 fragment = fragment[:-1]
-                suffix = u'.' + suffix
+                suffix = "." + suffix
                 continue
 
             # Nothing matched, so we're done
@@ -344,20 +389,30 @@ class LinkifyFilter(Filter):
 
     def handle_links(self, src_iter):
         """Handle links in character tokens"""
+        in_a = False  # happens, if parse_email=True and if a mail was found
         for token in src_iter:
-            if token['type'] == 'Characters':
-                text = token['data']
+            if in_a:
+                if token["type"] == "EndTag" and token["name"] == "a":
+                    in_a = False
+                yield token
+                continue
+            elif token["type"] == "StartTag" and token["name"] == "a":
+                in_a = True
+                yield token
+                continue
+            if token["type"] == "Characters":
+                text = token["data"]
                 new_tokens = []
                 end = 0
 
                 for match in self.url_re.finditer(text):
                     if match.start() > end:
                         new_tokens.append(
-                            {u'type': u'Characters', u'data': text[end:match.start()]}
+                            {"type": "Characters", "data": text[end : match.start()]}
                         )
 
                     url = match.group(0)
-                    prefix = suffix = ''
+                    prefix = suffix = ""
 
                     # Sometimes we pick up too much in the url match, so look for
                     # bits we should drop and remove them from the match
@@ -367,40 +422,35 @@ class LinkifyFilter(Filter):
                     if PROTO_RE.search(url):
                         href = url
                     else:
-                        href = u'http://%s' % url
+                        href = "http://%s" % url
 
-                    attrs = {
-                        (None, u'href'): href,
-                        u'_text': url
-                    }
+                    attrs = {(None, "href"): href, "_text": url}
                     attrs = self.apply_callbacks(attrs, True)
 
                     if attrs is None:
                         # Just add the text
                         new_tokens.append(
-                            {u'type': u'Characters', u'data': prefix + url + suffix}
+                            {"type": "Characters", "data": prefix + url + suffix}
                         )
 
                     else:
                         # Add the "a" tag!
                         if prefix:
-                            new_tokens.append(
-                                {u'type': u'Characters', u'data': prefix}
-                            )
+                            new_tokens.append({"type": "Characters", "data": prefix})
 
-                        _text = attrs.pop(u'_text', '')
+                        _text = attrs.pop("_text", "")
                         attrs = alphabetize_attributes(attrs)
 
-                        new_tokens.extend([
-                            {u'type': u'StartTag', u'name': u'a', u'data': attrs},
-                            {u'type': u'Characters', u'data': force_unicode(_text)},
-                            {u'type': u'EndTag', u'name': 'a'},
-                        ])
+                        new_tokens.extend(
+                            [
+                                {"type": "StartTag", "name": "a", "data": attrs},
+                                {"type": "Characters", "data": str(_text)},
+                                {"type": "EndTag", "name": "a"},
+                            ]
+                        )
 
                         if suffix:
-                            new_tokens.append(
-                                {u'type': u'Characters', u'data': suffix}
-                            )
+                            new_tokens.append({"type": "Characters", "data": suffix})
 
                     end = match.end()
 
@@ -408,7 +458,7 @@ class LinkifyFilter(Filter):
                     # Yield the adjusted set of tokens and then continue
                     # through the loop
                     if end < len(text):
-                        new_tokens.append({u'type': u'Characters', u'data': text[end:]})
+                        new_tokens.append({"type": "Characters", "data": text[end:]})
 
                     for new_token in new_tokens:
                         yield new_token
@@ -427,23 +477,23 @@ class LinkifyFilter(Filter):
 
         """
         a_token = token_buffer[0]
-        if a_token['data']:
-            attrs = a_token['data']
+        if a_token["data"]:
+            attrs = a_token["data"]
         else:
             attrs = {}
         text = self.extract_character_data(token_buffer)
-        attrs['_text'] = text
+        attrs["_text"] = text
 
         attrs = self.apply_callbacks(attrs, False)
 
         if attrs is None:
             # We're dropping the "a" tag and everything else and replacing
             # it with character data. So emit that token.
-            yield {'type': 'Characters', 'data': text}
+            yield {"type": "Characters", "data": text}
 
         else:
-            new_text = attrs.pop('_text', '')
-            a_token['data'] = alphabetize_attributes(attrs)
+            new_text = attrs.pop("_text", "")
+            a_token["data"] = alphabetize_attributes(attrs)
 
             if text == new_text:
                 # The callbacks didn't change the text, so we yield the new "a"
@@ -457,7 +507,7 @@ class LinkifyFilter(Filter):
                 # all the tokens between the start and end "a" tags and replace
                 # it with the new text
                 yield a_token
-                yield {'type': 'Characters', 'data': force_unicode(new_text)}
+                yield {"type": "Characters", "data": str(new_text)}
                 yield token_buffer[-1]
 
     def __iter__(self):
@@ -470,7 +520,7 @@ class LinkifyFilter(Filter):
             if in_a:
                 # Handle the case where we're in an "a" tag--we want to buffer tokens
                 # until we hit an end "a" tag.
-                if token['type'] == 'EndTag' and token['name'] == 'a':
+                if token["type"] == "EndTag" and token["name"] == "a":
                     # Add the end tag to the token buffer and then handle them
                     # and yield anything returned
                     token_buffer.append(token)
@@ -481,19 +531,17 @@ class LinkifyFilter(Filter):
                     # the tokens we're going to yield
                     in_a = False
                     token_buffer = []
-                    continue
-
                 else:
                     token_buffer.append(token)
-                    continue
+                continue
 
-            elif token['type'] in ['StartTag', 'EmptyTag']:
-                if token['name'] in self.skip_tags:
+            if token["type"] in ["StartTag", "EmptyTag"]:
+                if token["name"] in self.skip_tags:
                     # Skip tags start a "special mode" where we don't linkify
                     # anything until the end tag.
-                    in_skip_tag = token['name']
+                    in_skip_tag = token["name"]
 
-                elif token['name'] == 'a':
+                elif token["name"] == "a":
                     # The "a" tag is special--we switch to a slurp mode and
                     # slurp all the tokens until the end "a" tag and then
                     # figure out what to do with them there.
@@ -507,10 +555,10 @@ class LinkifyFilter(Filter):
             elif in_skip_tag and self.skip_tags:
                 # NOTE(willkg): We put this clause here since in_a and
                 # switching in and out of in_a takes precedence.
-                if token['type'] == 'EndTag' and token['name'] == in_skip_tag:
+                if token["type"] == "EndTag" and token["name"] == in_skip_tag:
                     in_skip_tag = None
 
-            elif not in_a and not in_skip_tag and token['type'] == 'Characters':
+            elif not in_a and not in_skip_tag and token["type"] == "Characters":
                 new_stream = iter([token])
                 if self.parse_email:
                     new_stream = self.handle_email_addresses(new_stream)
