@@ -9,7 +9,7 @@ def _attr_key(attr):
     ``None`` to an empty string.
 
     """
-    key = (attr[0][0] or ''), attr[0][1]
+    key = (attr[0][0] or ""), attr[0][1]
     return key
 
 
@@ -18,6 +18,4 @@ def alphabetize_attributes(attrs):
     if not attrs:
         return attrs
 
-    return OrderedDict(
-        [(k, v) for k, v in sorted(attrs.items(), key=_attr_key)]
-    )
+    return OrderedDict([(k, v) for k, v in sorted(attrs.items(), key=_attr_key)])
